import abc
import logging
from pathlib import Path
from typing import Callable, Dict, Generic, List, Optional, Tuple, Type, TypeVar

import matplotlib.pyplot as plt
import numpy as np
from opentuner.measurement.interface import MeasurementInterface
from opentuner.search.manipulator import ConfigurationManipulator, EnumParameter

from ._logging import PathLike, override_opentuner_config
from ._pareto import is_pareto_efficient

msg_logger = logging.getLogger(__name__)
KnobsT = Dict[str, str]
TunerConfigT = Dict[int, int]


class ApproxKnob:
    def __init__(self, name: str, **kwargs):
        self.name = name
        self.kwargs = kwargs

    def coexists_with(self, other: "ApproxKnob") -> bool:
        return False

    def __repr__(self):
        return f'Knob"{self.name}"({self.kwargs})'


class ApproxApp(abc.ABC):
    """Generic approximable application with operator & knob enumeration,
    and measures its own QoS and performance given a configuration.

    Parameters
    ----------
    op_knobs:
        a mapping from each operator (identified by str) to a list of applicable knobs.
    """

    def __init__(self, op_knobs: Dict[str, List[ApproxKnob]]) -> None:
        super().__init__()
        self.op_knobs = op_knobs
        # Also modifies self.op_knobs in place.
        self.baseline_knob = self._check_get_baseline_knob_(self.op_knobs)

    @abc.abstractmethod
    def measure_qos_perf(
        self, with_approxes: KnobsT, is_test: bool
    ) -> Tuple[float, float]:
        pass

    def get_tuner(self) -> "ApproxTuner":
        """We implement this function. Sets up an ApproxTuner instance
        which the user can directly call `tune()` on with opentuner parameters."""
        return ApproxTuner(self)

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Name of application. Acts as an identifier in many places, so
        the user should try to make it unique."""
        return ""

    @property
    def ops(self) -> List[str]:
        return list(self.op_knobs)

    @property
    def knobs(self) -> List[ApproxKnob]:
        knob_sets = [set(knobs) for knobs in self.op_knobs.values()]
        return list(set.union(*knob_sets))

    @staticmethod
    def _check_get_baseline_knob_(
        op_knobs: Dict[str, List[ApproxKnob]]
    ) -> "BaselineKnob":
        # Modifies op_knobs inplace.
        # Find the baseline knob if the user has one, or get a default one
        knob_sets = [set(knobs) for knobs in op_knobs.values()]
        knobs = list(set.union(*knob_sets))
        baselines = set(k for k in knobs if isinstance(k, BaselineKnob))
        if len(baselines) > 1:
            raise ValueError(f"Found multiple baseline knobs in op_knobs: {baselines}")
        if baselines:
            (baseline_knob,) = baselines
        else:
            baseline_knob = BaselineKnob()
        # Start checking if each layer has the baseline knob
        for knobs in op_knobs.values():
            if baseline_knob not in set(knobs):
                knobs.append(baseline_knob)
        return baseline_knob

    def add_baseline_to_knobs(self, approxes: KnobsT):
        approxes = approxes.copy()
        for op_name in self.ops:
            if op_name not in approxes:
                approxes[op_name] = self.baseline_knob.name
        return approxes


class BaselineKnob(ApproxKnob):
    def __init__(self, name: str = "__baseline__"):
        super().__init__(name)


class Config:
    def __init__(
        self, qos: float, perf: float, knobs: KnobsT, test_qos: Optional[float] = None
    ) -> None:
        self.qos = qos
        self.perf = perf
        self.knobs = knobs
        self.test_qos: Optional[float] = test_qos


T = TypeVar("T", bound=Config)


# IOpenTuner is generic over the type of the config
# So that the user can use custom Config inherited from Config
# (in which case they need to override `get_all_configs_from_db`).
class ApproxTuner(Generic[T]):
    def __init__(self, app: ApproxApp) -> None:
        self.app = app
        self._tuned = False
        self.all_configs = []
        self.kept_configs = []
        self.best_configs = []
        # The following will be filled after self.tune() is called
        self.keep_threshold = None

    @property
    def tuned(self) -> bool:
        return self._tuned

    def tune(
        self,
        max_iter: int,
        qos_tuner_threshold: float,
        qos_keep_threshold: Optional[float] = None,
        is_threshold_relative: bool = False,
        take_best_n: Optional[int] = None,
        test_configs: bool = True,
        **kwargs
        # TODO: more parameters + opentuner param forwarding
    ) -> List[T]:
        from opentuner.tuningrunmain import TuningRunMain

        from ._dbloader import read_opentuner_db

        n_ops, n_knobs = len(self.app.ops), len(self.app.knobs)
        msg_logger.info(
            "Started tuning app %s with %d ops and %d unique knob types",
            self.app.name,
            n_ops,
            n_knobs,
        )
        msg_logger.info("At most %d iterations", max_iter)
        opentuner_args = opentuner_default_args()
        tuner = self._get_tuner_interface(
            opentuner_args,
            max_iter,
            qos_tuner_threshold,
            qos_keep_threshold,
            is_threshold_relative,
            self._get_app_kwargs(**kwargs),
        )
        assert self.keep_threshold is not None
        trm = TuningRunMain(tuner, opentuner_args)
        # TuningRunMain.__init__ initializes its own logger, so we'll override it and use ours
        override_opentuner_config()
        msg_logger.info(
            "Estimated size of search space: %d", trm.manipulator.search_space_size()
        )
        # A little bit of hack to get the _real_ progress when duplicated configs exist
        tuner.set_progress_getter(lambda: trm.search_driver.test_count)
        # This is where opentuner runs
        trm.main()
        # Parse and store results
        self._tuned = True
        config_ty = self._get_config_class()
        self.all_configs = [
            config_ty(result.accuracy, result.time, configuration.data)
            for result, configuration in read_opentuner_db(opentuner_args.database)
        ]
        self.kept_configs = [
            cfg for cfg in self.all_configs if cfg.qos > self.keep_threshold
        ]
        self.best_configs = self.take_best_configs(self.kept_configs, take_best_n)
        msg_logger.info(
            "Tuning finished with %d configs in total, "
            "%d configs above keeping threshold, "
            "and %d configs selected on tradeoff curve",
            len(self.all_configs),
            len(self.kept_configs),
            len(self.best_configs),
        )
        if test_configs:
            msg_logger.info("Checking configurations on test inputs")
            self.test_configs_(self.best_configs)
        return self.best_configs

    def test_configs_(self, configs: List[T]):
        from tqdm import tqdm

        for cfg in tqdm(configs, leave=False):
            cfg: T
            if cfg.test_qos is not None:
                continue
            cfg.test_qos, _ = self.app.measure_qos_perf(cfg.knobs, True)
            msg_logger.debug(f"Calibration: {cfg.qos} (mean) -> {cfg.test_qos} (mean)")

    @staticmethod
    def take_best_configs(configs: List[T], n: Optional[int] = None) -> List[T]:
        points = np.array([[c.perf, c.qos] for c in configs])
        taken_idx = is_pareto_efficient(points, take_n=n)
        return [configs[i] for i in taken_idx]

    def dump_configs(self, filepath: PathLike):
        import os

        from jsonpickle import encode

        if not self.tuned:
            raise RuntimeError(
                f"No tuning session has been run; call self.tune() first."
            )
        filepath = Path(filepath)
        os.makedirs(filepath.parent, exist_ok=True)
        with filepath.open("w") as f:
            f.write(encode(self.kept_configs, indent=2))

    def plot_configs(self) -> plt.Figure:
        if not self.tuned:
            raise RuntimeError(
                f"No tuning session has been run; call self.tune() first."
            )

        _, perf = self.app.measure_qos_perf({}, False)
        fig, ax = plt.subplots()
        confs = self.kept_configs
        if not confs:
            return fig
        qos_speedup = [(c.qos, perf / c.perf) for c in confs]
        qoses, speedups = zip(*sorted(qos_speedup, key=lambda p: p[0]))
        ax.plot(qoses, speedups)
        ax.scatter(qoses, speedups)
        ax.set_xlabel("qos")
        ax.set_ylabel("speedup")
        return fig

    def _get_tuner_interface(
        self,
        opentuner_args,
        max_iter: int,
        qos_tuner_threshold: float,
        qos_keep_threshold: Optional[float],
        is_threshold_relative: bool,
        app_kwargs: dict,
    ) -> "TunerInterface":
        # By default, keep_threshold == tuner_threshold
        self.keep_threshold = qos_keep_threshold or qos_tuner_threshold
        if is_threshold_relative:
            baseline_qos, _ = self.app.measure_qos_perf({}, False)
            qos_tuner_threshold = baseline_qos - qos_tuner_threshold
            self.keep_threshold = baseline_qos - self.keep_threshold
        opentuner_args.test_limit = max_iter
        msg_logger.info(
            "Tuner QoS threshold: %f; keeping configurations with QoS >= %f",
            qos_tuner_threshold,
            self.keep_threshold,
        )
        return TunerInterface(
            opentuner_args,
            self.app,
            qos_tuner_threshold,
            self.keep_threshold,
            max_iter,
            **app_kwargs,
        )

    def _get_app_kwargs(self, **kwargs):
        return {}

    @classmethod
    def _get_config_class(cls) -> Type[Config]:
        return Config


def opentuner_default_args():
    from opentuner import default_argparser

    args = default_argparser().parse_args([])
    args.no_dups = True  # Don't print duplicated config warnings
    return args


class TunerInterface(MeasurementInterface):
    def __init__(
        self,
        args,
        app: ApproxApp,
        tuner_thres: float,
        keep_thres: float,
        test_limit: int,
        **app_kwargs,
    ):
        from opentuner.measurement.inputmanager import FixedInputManager
        from opentuner.search.objective import ThresholdAccuracyMinimizeTime
        from tqdm import tqdm

        self.app = app
        self.tune_thres = tuner_thres
        self.keep_thres = keep_thres
        self.pbar = tqdm(total=test_limit, leave=False)
        self.app_kwargs = app_kwargs

        # tune_thres can come in as np.float64 and opentuner doesn't like that
        objective = ThresholdAccuracyMinimizeTime(float(tuner_thres))
        input_manager = FixedInputManager(size=len(self.app.op_knobs))
        super(TunerInterface, self).__init__(
            args,
            program_name=self.app.name,
            input_manager=input_manager,
            objective=objective,
        )

    def set_progress_getter(self, getter: Callable[[], int]):
        self.progress_getter = getter

    def manipulator(self) -> ConfigurationManipulator:
        """Define the search space by creating a ConfigurationManipulator."""
        manipulator = ConfigurationManipulator()
        for op, knobs in self.app.op_knobs.items():
            knob_names = [knob.name for knob in knobs]
            manipulator.add_parameter(EnumParameter(op, knob_names))
        return manipulator

    def run(self, desired_result, input_, limit):
        """Run a given configuration then return performance and accuracy."""
        from opentuner.resultsdb.models import Result

        cfg = desired_result.configuration.data
        qos, perf = self.app.measure_qos_perf(cfg, False, **self.app_kwargs)
        # Print a debug message for each config in tuning and keep threshold
        self.print_debug_config(qos, perf)
        self.pbar.update(self.progress_getter() - self.pbar.n)
        return Result(time=perf, accuracy=qos)

    def save_final_config(self, config):
        self.pbar.close()

    def print_debug_config(self, qos: float, perf: float):
        gt_tune, gt_keep = qos > self.tune_thres, qos > self.keep_thres
        if not gt_tune and not gt_keep:
            return
        if gt_tune and not gt_keep:
            kind = "tuning"
        elif gt_keep and not gt_tune:
            kind = "keep"
        else:
            kind = "tuning and keep"
        msg_logger.debug(
            f"Found config within {kind} threshold: QoS = {qos}, perf = {perf}"
        )
