import abc
import json
import logging
import pickle
from pathlib import Path
from typing import Callable, Dict, Iterator, List, Optional, Tuple, Type, Union

import numpy as np
import pandas as pd
import torch

from ._logging import PathLike
from .approxapp import ApproxApp, ApproxKnob, ApproxTuner, Config, KnobsT

msg_logger = logging.getLogger(__name__)


class ModeledApp(ApproxApp, abc.ABC):
    """Approximable application that inherits at least 1 interface for performance/QoS modeling.

    It's invalid to inherit from this class without also implementing at least 1 interface
    provided in this set of API;
    for non-modeling application, inherit from `ApproxApp` instead.
    """

    def __init__(self, op_knobs: Dict[str, List[ApproxKnob]]) -> None:
        super().__init__(op_knobs)
        models = self.get_models()
        self._name_to_model = {m.name: m for m in models}
        if len(self._name_to_model) != len(models):
            raise ValueError("Name conflict in models")
        self._perf_models = {
            model.name: model for model in models if isinstance(model, IPerfModel)
        }
        self._qos_models = {
            model.name: model for model in models if isinstance(model, IQoSModel)
        }

    @abc.abstractmethod
    def get_models(self) -> List[Union["IPerfModel", "IQoSModel"]]:
        """Get QoS/Performance prediction models for this application."""
        pass

    def empirical_measure_qos_perf(
        self, with_approxes: KnobsT, is_test: bool
    ) -> Tuple[float, float]:
        """Measures QoS and performance by running the program with approximation.

        An implementation is not necessary if empirical measurement is never intended.
        """
        raise NotImplementedError()

    def measure_qos_perf(
        self,
        with_approxes: KnobsT,
        is_test: bool,
        qos_model: str = "none",
        perf_model: str = "none",
    ) -> Tuple[float, float]:
        """We provide this with the right qos and perf function.

        Empirical measurement will be called once if either `perf_model` or `qos_model`
        is "none", otherwise only use model indicated by model name.
        """
        # Testset measurement is always empirical
        if is_test:
            return self.empirical_measure_qos_perf(with_approxes, is_test)
        # Run empirical measurement once if either perf or qos needs it
        qos, perf = None, None
        if qos_model == "none" or perf_model == "none":
            qos, perf = self.empirical_measure_qos_perf(with_approxes, is_test)
        # If we're asked to use some qos_model, overwrite `qos` value
        # even if we already get it from empirical measure (i.e., even if perf_model == "none")
        if qos_model != "none":
            if qos_model not in self._qos_models:
                raise ValueError(
                    f'"{qos_model}" is an invalid value for qos_model '
                    f"(choose from {list(self._qos_models.keys())})"
                )
            qos = self._qos_models[qos_model].measure_qos(with_approxes)
        # Same goes for perf
        if perf_model != "none":
            if perf_model not in self._perf_models:
                raise ValueError(
                    f'"{perf_model}" is an invalid value for perf_model '
                    f"(choose from {list(self._perf_models.keys())})"
                )
            perf = self._perf_models[perf_model].measure_perf(with_approxes)
        assert type(qos) is float and type(perf) is float
        return qos, perf

    def get_tuner(self) -> "ApproxModeledTuner":
        return ApproxModeledTuner(self)

    def init_model(self, model_name: str):
        self._name_to_model[model_name]._init()


class IPerfModel(abc.ABC):
    """Abstract base class for models that provide performance prediction."""

    def __init__(self) -> None:
        self._inited = False

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Name of model."""
        pass

    @abc.abstractmethod
    def measure_perf(self, with_approxes: KnobsT) -> float:
        """Predict the performance of application."""
        pass

    def _init(self):
        """Initialize the model before the first prediction task (profiling, etc.)"""
        self._inited = True


class IQoSModel(abc.ABC):
    """Abstract base class for models that provide QoS prediction."""

    def __init__(self) -> None:
        self._inited = False

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Name of model."""
        pass

    @abc.abstractmethod
    def measure_qos(self, with_approxes: KnobsT) -> float:
        """Predict the qos of application."""
        pass

    def _init(self):
        """Initialize the model before the first prediction task (profiling, etc.)"""
        self._inited = True


class LinearPerfModel(IPerfModel):
    """Weighted linear performance predictor based on cost of each operator."""

    def __init__(
        self,
        app: ModeledApp,
        op_costs: Dict[str, float],
        knob_speedups: Dict[str, float],
    ) -> None:
        import numpy as np
        import pandas as pd

        super().__init__()
        self.app = app
        knob_cost_factor_v = 1 / np.array(list(knob_speedups.values()))
        layer_cost_v = np.array(list(op_costs.values()))
        costs = np.outer(layer_cost_v, knob_cost_factor_v)
        self.cost_df = pd.DataFrame(
            costs, index=op_costs.keys(), columns=knob_speedups.keys(), dtype=float
        )

    @property
    def name(self) -> str:
        return "perf_linear"

    def measure_perf(self, with_approxes: KnobsT) -> float:
        """We implement this using a weighted linear performance model."""
        with_approxes = self.app.add_baseline_to_knobs(with_approxes)
        return float(
            sum(self.cost_df.loc[layer, knob] for layer, knob in with_approxes.items())
        )


class QoSModelP1(IQoSModel):
    """QoS model `P1` in ApproxTuner.

    tensor_output_getter: Run the tensor-based application with config `with_approxes` applied,
        and return a single tensor result.

        Note that while we require the return value to be a PyTorch tensor,
        user is free to implement this on non-PyTorch applications.

    qos_metric: Compute a Quality of Service level from the tensor output of application
    """

    def __init__(
        self,
        app: ModeledApp,
        tensor_output_getter: Callable[[KnobsT], torch.Tensor],
        qos_metric: Callable[[torch.Tensor], float],
        storage: PathLike = None,
    ) -> None:
        super().__init__()
        self.app = app
        self.output_f = tensor_output_getter
        self.qos_metric = qos_metric
        self.storage = Path(storage) if storage else None
        self.delta_tensors = {
            op: {k.name: None for k in self.app.knobs} for op in self.app.ops
        }
        self.baseline_tensor = None

    @property
    def name(self) -> str:
        return "qos_p1"

    def measure_qos(self, with_approxes: KnobsT) -> float:
        """Implementation of model."""
        assert self.baseline_tensor is not None
        with_approxes = self.app.add_baseline_to_knobs(with_approxes)
        delta_tensors = np.array(
            [self.delta_tensors[op][knob] for op, knob in with_approxes.items()]
        )
        ret = delta_tensors.sum() + self.baseline_tensor
        return float(self.qos_metric(ret))

    def _init(self):
        dt = self.delta_tensors
        btensor = self.baseline_tensor = self.output_f({})
        if self.storage and self.storage.is_file():
            for op, knob, delta_tensor in self._load(self.storage):
                dt[op][knob] = delta_tensor
        for op, knob in barred_ravel_knobs(self.app):
            if dt[op][knob] is not None:
                continue
            delta_tensor = self.output_f({op: knob}) - btensor
            dt[op][knob] = delta_tensor
            self._try_append_save(self.storage, op, knob, delta_tensor)
        super()._init()

    def _load(self, path: Path) -> Iterator[Tuple[str, str, torch.Tensor]]:
        msg_logger.info(f"Model {self.name} found saved model at {path}")
        with path.open("rb") as f:
            while True:
                try:
                    op_name, knob_name, tensor = pickle.load(f)
                    yield op_name, knob_name, tensor
                except EOFError:
                    return

    @staticmethod
    def _try_append_save(
        path: Optional[Path], op_name: str, knob_name: str, tensor: torch.Tensor
    ):
        import os

        if not path:
            return
        if not path.parent.is_dir():
            os.makedirs(path.parent)
        with path.open("ab") as f:
            pickle.dump((op_name, knob_name, tensor), f)


class QoSModelP2(IQoSModel):
    """QoS model `P2` in ApproxTuner."""

    def __init__(self, app: ModeledApp, storage: PathLike = None) -> None:
        super().__init__()
        self.app = app
        self.storage = Path(storage) if storage else None
        self.qos_df = None
        self.baseline_qos = None

    @property
    def name(self) -> str:
        return "qos_p2"

    def _empirical_measure_qos(self, with_approxes: KnobsT) -> float:
        """An internal QoS-measuring method.

        The point is P2 queries some QoS results and caches them before tuning starts,
        and then defines a `measure_qos` that doesn't run the application during tuning
        (to reduce overhead).
        """
        qos, _ = self.app.empirical_measure_qos_perf(with_approxes, False)
        return qos

    def measure_qos(self, with_approxes: KnobsT) -> float:
        assert self.baseline_qos is not None and self.qos_df is not None
        with_approxes = self.app.add_baseline_to_knobs(with_approxes)
        delta_qoses = (
            np.array([self.qos_df.loc[kv] for kv in with_approxes.items()])
            - self.baseline_qos
        )
        ret = delta_qoses.sum() + self.baseline_qos
        assert not np.isnan(ret)
        return float(ret)

    def _init(self):
        if self.storage and self.storage.is_file():
            self.qos_df, self.baseline_qos = self._load(self.storage)
        else:
            knob_names = [k.name for k in self.app.knobs]
            self.qos_df = pd.DataFrame(index=self.app.ops, columns=knob_names)
            self.qos_df = self.qos_df.where(pd.notnull(self.qos_df), None)
            self.baseline_qos = self._empirical_measure_qos({})
        df = self.qos_df
        for op, knob in barred_ravel_knobs(self.app):
            if df.loc[op, knob] is not None:
                continue
            df.loc[op, knob] = self._empirical_measure_qos({op: knob})
        if self.storage and not self.storage.is_file():
            self._save(self.storage)
        super()._init()

    def _load(self, path: Path) -> Tuple[pd.DataFrame, float]:
        with path.open() as f:
            data = json.load(f)
        df = pd.DataFrame(data["df"])
        baseline_qos = float(data["bqos"])
        if "app_name" in data:
            name = data["app_name"]
            if self.app.name != name:
                msg_logger.error(
                    f'Profile at {path} belongs to app "{name}" '
                    f"while our app is {self.app.name}"
                )
        else:
            msg_logger.warning("Loaded profile does not have app name identifier")
        msg_logger.info(f"Model {self.name} loaded saved model at {path}")
        return df, baseline_qos

    def _save(self, path: Path):
        import os

        if not path.parent.is_dir():
            os.makedirs(path.parent)
        with path.open("w") as f:
            json.dump(
                {
                    "app_name": self.app.name,
                    "df": self.qos_df.to_dict(),
                    "bqos": self.baseline_qos,
                },
                f,
            )


class ValConfig(Config):
    def __init__(
        self,
        qos: float,
        perf: float,
        knobs: KnobsT,
        test_qos: Optional[float] = None,
        validated_qos: Optional[float] = None,
    ) -> None:
        super().__init__(qos, perf, knobs, test_qos)
        self.validated_qos = validated_qos


class ApproxModeledTuner(ApproxTuner):
    app: ModeledApp

    def tune(
        self,
        max_iter: int,
        qos_tuner_threshold: float,
        qos_keep_threshold: Optional[float] = None,
        is_threshold_relative: bool = False,
        take_best_n: Optional[int] = None,
        test_configs: bool = True,
        validate_configs: Optional[bool] = None,
        perf_model: str = "none",
        qos_model: str = "none",
    ) -> List[ValConfig]:
        qos_desc = (
            "no model for qos" if qos_model == "none" else f'qos model "{qos_model}"'
        )
        perf_desc = (
            "no model for performance"
            if perf_model == "none"
            else f'performance model "{perf_model}"'
        )
        msg_logger.info("Starting tuning with %s and %s", qos_desc, perf_desc)
        if qos_model != "none":
            msg_logger.info("Initializing qos model %s", qos_model)
            self.app.init_model(qos_model)
        if perf_model != "none":
            msg_logger.info("Initializing performance model %s", perf_model)
            self.app.init_model(perf_model)
        ret = super().tune(
            max_iter=max_iter,
            qos_tuner_threshold=qos_tuner_threshold,
            qos_keep_threshold=qos_keep_threshold,
            is_threshold_relative=is_threshold_relative,
            take_best_n=take_best_n,
            test_configs=test_configs,
            perf_model=perf_model,
            qos_model=qos_model,
        )
        if validate_configs is None and qos_model != "none":
            msg_logger.info(
                'Validating configurations due to using qos model "%s"', qos_model
            )
            self.validate_configs_(self.best_configs)
        elif validate_configs:
            msg_logger.info("Validating configurations as user requested")
            self.validate_configs_(self.best_configs)
        return ret

    def validate_configs_(self, configs: List[ValConfig]):
        from tqdm import tqdm

        for cfg in tqdm(configs, leave=False):
            cfg: ValConfig
            if cfg.validated_qos is not None:
                continue
            cfg.validated_qos, _ = self.app.measure_qos_perf(cfg.knobs, False)
            msg_logger.debug(f"Validation: {cfg.qos} (mean) -> {cfg.test_qos} (mean)")

    def _get_app_kwargs(self, perf_model: str, qos_model: str):
        return {"perf_model": perf_model, "qos_model": qos_model}

    @classmethod
    def _get_config_class(cls) -> Type[Config]:
        return ValConfig


def barred_ravel_knobs(app: ApproxApp) -> Iterator[Tuple[str, str]]:
    """Flattens op_knobs of app to a list of (layer, knob) pairs while showing 2 levels of
    progress bar."""

    from tqdm import tqdm

    bar1 = tqdm(app.op_knobs.items(), leave=None)
    for op_name, knobs in bar1:
        bar1.set_postfix(op=op_name)
        bar2 = tqdm(knobs, leave=None)
        for knob in bar2:
            bar2.set_postfix(knob=knob.name)
            yield op_name, knob.name
