import abc
from pathlib import Path
from typing import Any, Callable, List, Optional, Set, Tuple, Union

import numpy as np
import torch
from torch.nn import Module
from torch.utils.data.dataloader import DataLoader

from ._logging import PathLike
from .approxapp import ApproxKnob, BaselineKnob, KnobsT
from .modeledapp import (IPerfModel, IQoSModel, LinearPerfModel, ModeledApp,
                         QoSModelP1, QoSModelP2)
from .torchutil import ModuleIndexer, get_summary, move_to_device_recursively


class TorchApproxKnob(ApproxKnob):
    """Defines an approximation knob that knows
    its own expected speedup ratio and what Modules it can apply to,
    and can be applied to a torch.nn.Module to return an approximated Module."""

    @property
    @abc.abstractmethod
    def deterministic(self) -> bool:
        """Returns true if approx knob does not contain randomness."""
        pass

    @property
    @abc.abstractmethod
    def expected_speedup(self) -> float:
        pass

    @abc.abstractmethod
    def is_applicable(self, op: Module) -> bool:
        pass

    @abc.abstractmethod
    def apply(self, op: Module) -> Module:
        """Applies knob to `module` and returns an approximated `module`."""
        pass


_default_device = f"cuda" if torch.cuda.is_available() else "cpu"


class TorchApp(ModeledApp, abc.ABC):
    r"""Adaptor for approximable PyTorch Modules with tensor output.

    A TorchApp stores the PyTorch Module, datasets for tuning and calibration,
    set of available TorchApproxKnob each of which may be applied to some layer in the Module,
    and the quality of service (QoS) metric of application (e.g., accuracy).
    It provides empirical tuning and predictive tuning capability (see `TorchApp.tune()`),

    Parameters
    ----------
    app_name:
        Name of the application, which is used as an identifier in tuning sessions, etc.
    module:
        The PyTorch module to tune.
    tune_dataloader:
        A dataset to use as inputs to module during tuning. (PyTorch DataLoader is conceptually
        an enumerable, batched dataset.)
    test_dataloader:
        A input dataset used for QoS testing (see `test_config` parameter of `ApproxModeledTuner.tune`).
    knobs:
        A set of knobs to be considered. Each knob has an `is_applicable()` method
        which is used to determine which layer it can apply to.
    tensor_to_qos:
        QoS metric function which computes QoS from the module's output.
    combine_qos:
        A function to combine each batch's QoS into one value.
        When QoS is accuracy this will most likely be `mean()` (which is the default).
    device:
        The device to store module and perform inference on. By default is "cuda"
        if CUDA is available, otherwise "cpu".
    model_storage_folder:
        A folder to store the serialized QoS models into.
        `QoSModelP1` will be serialized into `model_storage_folder / "p1.pkl"`,
        and `QoSModelP2` into `model_storage_folder / "p2.json"`.
        See `QoSModelP1` and `QoSModelP2` for details.

    Attributes
    ----------
    """

    def __init__(
        self,
        app_name: str,
        module: Module,
        tune_dataloader: DataLoader,
        test_dataloader: DataLoader,
        knobs: Set[TorchApproxKnob],
        tensor_to_qos: Callable[[torch.Tensor, Any], float],
        combine_qos: Callable[[np.ndarray], float] = np.mean,
        device: Union[torch.device, str] = _default_device,
        model_storage_folder: Optional[PathLike] = None,
    ) -> None:
        self.app_name = app_name
        self.module = module
        self.tune_loader = tune_dataloader
        self.test_loader = test_dataloader
        self.name_to_knob = {k.name: k for k in self._check_baseline_knob(knobs)}
        self.tensor_to_qos = tensor_to_qos
        self.combine_qos = combine_qos
        self.device = device
        self.model_storage = (
            Path(model_storage_folder) if model_storage_folder else None
        )

        self.module = self.module.to(device)
        self.midx = ModuleIndexer(module)
        self._op_costs = {}
        op_knobs = {}
        self._knob_speedups = {k.name: k.expected_speedup for k in knobs}
        modules = self.midx.name_to_module
        summary = get_summary(self.module, (self._sample_input(),))
        for op_name, op in modules.items():
            this_knobs = [
                knob for knob in self.name_to_knob.values() if knob.is_applicable(op)
            ]
            assert this_knobs
            op_knobs[op_name] = this_knobs
            self._op_costs[op_name] = summary.loc[op_name, "flops"]

        # Init parent class last
        super().__init__(op_knobs)

    @property
    def name(self) -> str:
        """Returns the name of application."""
        return self.app_name

    def get_models(self) -> List[Union[IPerfModel, IQoSModel]]:
        """Returns a list of predictive tuning models.

        TorchApp in particular derives 1 performance model (LinearPerfModel)
        and 2 QoS models (QoSModelP1, QoSModelP2) automatically.
        """

        def batched_valset_qos(tensor_output: torch.Tensor):
            dataset_len = len(self.tune_loader.dataset)
            assert len(tensor_output) == dataset_len
            begin = 0
            qoses = []
            for _, target in self.tune_loader:
                end = begin + len(target)
                target = move_to_device_recursively(target, self.device)
                qos = self.tensor_to_qos(tensor_output[begin:end], target)
                qoses.append(qos)
            return self.combine_qos(np.array(qoses))

        p1_storage = self.model_storage / "p1.pkl" if self.model_storage else None
        p2_storage = self.model_storage / "p2.json" if self.model_storage else None
        return [
            LinearPerfModel(self, self._op_costs, self._knob_speedups),
            QoSModelP1(
                self, self._get_raw_output_valset, batched_valset_qos, p1_storage
            ),
            QoSModelP2(self, p2_storage),
        ]

    @torch.no_grad()
    def empirical_measure_qos_perf(
        self, with_approxes: KnobsT, is_test: bool
    ) -> Tuple[float, float]:
        """Measure the QoS and performance of Module with given approximation
        empirically (i.e., by running the Module on the dataset)."""

        from time import time_ns

        dataloader = self.test_loader if is_test else self.tune_loader
        approxed = self._apply_knobs(with_approxes)
        qoses = []

        time_begin = time_ns() / (10 ** 9)
        for inputs, targets in dataloader:
            inputs = move_to_device_recursively(inputs, self.device)
            targets = move_to_device_recursively(targets, self.device)
            outputs = approxed(inputs)
            qoses.append(self.tensor_to_qos(outputs, targets))
        time_end = time_ns() / (10 ** 9)
        qos = float(self.combine_qos(np.array(qoses)))
        return qos, time_end - time_begin

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        module_class_name = type(self.module).__name__
        return (
            f'{class_name}"{self.name}"(module={module_class_name}, '
            f"num_op={len(self.op_knobs)}, num_knob={len(self.name_to_knob)})"
        )

    @torch.no_grad()
    def _get_raw_output_valset(self, with_approxes: KnobsT):
        approxed = self._apply_knobs(with_approxes)
        all_outputs = []
        for inputs, _ in self.tune_loader:
            inputs = move_to_device_recursively(inputs, self.device)
            outputs = approxed(inputs)
            all_outputs.append(outputs)
        return torch.cat(all_outputs, dim=0)

    @staticmethod
    def _check_baseline_knob(knobs: Set[TorchApproxKnob]) -> Set[TorchApproxKnob]:
        baselines = set(k for k in knobs if isinstance(k, TorchBaselineKnob))
        if len(baselines) > 1:
            raise ValueError(f"Found multiple baseline knobs in op_knobs: {baselines}")
        if not baselines:
            print("Adding baseline knob to knob set")
            knobs.add(TorchBaselineKnob())
        return knobs

    def _apply_knobs(self, knobs: KnobsT) -> Module:
        import copy

        module_indexer = copy.deepcopy(self.midx)
        for op_name, knob_name in knobs.items():
            knob = self.name_to_knob[knob_name]
            module_indexer[op_name] = knob.apply(module_indexer[op_name])
        return module_indexer.module

    def _sample_input(self):
        inputs, _ = next(iter(self.tune_loader))
        return inputs.to(self.device)


class TorchBaselineKnob(TorchApproxKnob, BaselineKnob):
    def __init__(self, name: str = "__baseline__"):
        super().__init__(name)

    @property
    def deterministic(self) -> bool:
        return True

    @property
    def expected_speedup(self) -> float:
        return 1.0

    def is_applicable(self, op: Module) -> bool:
        return True

    def apply(self, op: Module) -> Module:
        return op
